import networkx as nx


def actual(graph, sg):
    ids = graph.get_ids(stable=True)

    if sg is not None:
        # if sg > 0:
        # nx.set_node_attributes(graph.get_nx_graph(), name="tel", values={n: float(pwrlaw(sg, 1, 0, 1)) for n in ids})
        if sg == 0:
            nx.set_node_attributes(graph.get_nx_graph(), name="tel", values=1)
    nx.set_node_attributes(
        graph.get_nx_graph(),
        name="ind",
        values={n: graph.count_in_degree(n) for n in ids},
    )
    nx.set_node_attributes(
        graph.get_nx_graph(),
        name="out",
        values={n: graph.count_out_degree(n) for n in ids},
    )



import numpy as np
import networkx as nx

from pge.init.classes.graph import SGraph


class BasicGrowth:
    def __init__(self, graph, schema, deg, model_param):
        self.gr = graph
        self.deg = deg
        self.schema = schema
        self.param = model_param
        self.count = 0

    def choice(self, gr, sz):
        return []

    def new_edge_add(self, gr, attrs):
        return

    def new_node_add(self, graph, to, tp, attrs):
        return

    def prep(self, graph):
        return graph

    @staticmethod
    def make_copy(gr):
        if gr.directed():
            graph = nx.MultiDiGraph()
        else:
            graph = nx.MultiGraph()

        count = 0
        for edge in gr.get_edges():
            graph.add_edge(edge[0], edge[1], key="old-"+str(count))
            count += 1
        return SGraph(graph)

    def proceed(self, n, save=None, attr="cnt"):
        self.count = 0
        nw_graph = self.make_copy(self.gr)
        nx.set_node_attributes(nw_graph.get_nx_graph(), 0, name=attr)
        nx.set_edge_attributes(nw_graph.get_nx_graph(), 0, name=attr)
        nw_graph = self.new_load(nw_graph)

        count = self.gr.size()
        for _ in np.arange(n):
            print(_)
            if self.stop():
                break

            nw_graph = self.prep(nw_graph)
            new_node = np.random.choice(len(self.schema), p=self.schema)
            if new_node == 1:
                self.new_edge_add(nw_graph, (attr, _))
            else:
                self.new_node_add(nw_graph, str(count), new_node, (attr, _))
                count += 1
        if save is None:
            return nw_graph
        else:
            self.save(nw_graph, save)

    def save(self, gr, to):
        nx.write_graphml(gr.get_nx_graph(), to + ".graphml")

    def new_load(self, gr):
        return gr

    def stop(self):
        return False


class BDiGrowth(BasicGrowth):
    def new_node_add(self, graph, to, tp, attrs):
        if tp == 0:
            if self.deg[0][0] == "const":
                nodes = self.choice(graph, self.deg[0][1], tp="in")
            else:
                nodes = self.choice(graph, self.deg[0][0](self.deg[0][1]), tp="in")

            for node in nodes:
                graph.add_edge(to, node, str(self.count), {attrs[0]: attrs[1] + 1})
                self.count += 1
        else:
            if self.deg[1][0] == "const":
                nodes = self.choice(graph, self.deg[1][1], tp="out")
            else:
                nodes = self.choice(graph, self.deg[1][0](self.deg[1][1]), tp="out")

            for node in nodes:
                graph.add_edge(node, to, str(self.count), {attrs[0]: attrs[1] + 1})
                self.count += 1
        graph.set_attr(to, attrs[0], attrs[1] + 1)

    def new_edge_add(self, gr, attrs):
        node1, node2 = self.choice(gr, 1, tp="out")[0], self.choice(gr, 1, tp="in")[0]
        gr.add_edge(node1, node2, str(self.count), {attrs[0]: attrs[1] + 1})
        self.count += 1

    def choice(self, gr, sz, tp="in"):
        return []


class BUnGrowth(BasicGrowth):
    def new_node_add(self, graph, to, tp, attrs):
        if self.deg[0] == "const":
            nodes = self.choice(graph, self.deg[1])
        else:
            nodes = self.choice(graph, self.deg[0](self.deg[1]))

        for node in nodes:
            graph.add_edge(to, node, {attrs[0]: attrs[1] + 1})
        graph.set_attr(to, attrs[0], attrs[1] + 1)

    def new_edge_add(self, gr, attrs):
        nodes = self.choice(gr, 2)
        gr.add_edge(nodes[0], nodes[1], {attrs[0]: attrs[1] + 1})



import numpy as np

from pge.ranks.onlcore import DepNodeEx
from pge.ranks.onlcore import GravNodeEx


class SlideNodeEx(GravNodeEx, DepNodeEx):
    def __init__(self, eps, tree=True, quick=True):
        super().__init__(eps, quick)
        self.tree = tree

    def get_params(self, gr, c_d, frm, way):
        if self.tree:
            return DepNodeEx.get_params(self, gr, c_d, frm, way)
        else:
            return GravNodeEx.get_params(self, gr, c_d, frm, way)

    def get_ex_field(self, gr, root, frm, c):
        if self.tree:
            return DepNodeEx.get_ex_field(self, gr, root, frm, c)
        else:
            return GravNodeEx.get_ex_field(self, gr, root, frm, c)

    def get_ex(self, gr, frm, to):
        res = []

        for node in self.f:
            exc = gr.get_in_degrees(node)

            if exc.size == 0:
                continue

            nodes = gr.get_in_degrees(self.f, ex_i=gr.get_in_degrees(node), un=False)

            if nodes.size == 0:
                continue

            mx = np.max(gr.get_attributes(frm, exc))
            subs = np.where(gr.get_attributes(frm, nodes) <= mx, 1, 0)
            res.append(max([-exc.size*np.log(np.mean(subs)), 1]))

        if len(res) > 0:
            return 1/np.mean(res)
        else:
            return 0

    @staticmethod
    def save_param(gr, n, res, to):
        gr.set_attr(n, to, res)



import numpy as np
from random import choice, randint

from pge.model.spreading.spreader.fix_size import FixSpreadModel


class UniformFixGossip(FixSpreadModel):
    def iteration(self):
        ind = randint(0, self.ids.size-1)
        node = self.ids[ind]

        others = self.graph.get_out_degrees(node)

        if others.size != 0:
            u = choice(others)
            old = list(self.status[node])
            (self.status[node]).update(self.status[u])

            if self.received[ind] != len(self.status[node]):
                self.received[ind] = len(self.status[node])
                nws = np.setdiff1d(list(self.status[node]), old)
                self.messes[np.isin(self.ids, nws)] += 1
                return True
        return False



import networkx as nx
import numpy as np

from pge.model.growth.basic import BUnGrowth
from pge.model.growth.basic_fix import FixUnGrowth


class CAGrowth(BUnGrowth):
    def choice(self, graph, sz):
        cs = {
            k: v ** self.param[0] + self.param[1]
            for k, v in nx.clustering(graph.get_nx_graph()).items()
            if v is not np.NaN
        }
        probs = list(cs.values())
        probs = probs / np.sum(probs)
        return np.random.choice(list(cs.keys()), sz, replace=False, p=probs)


class CAFixGrowth(FixUnGrowth):
    def prep(self, graph):
        rs = nx.clustering(graph.get_nx_graph())
        graph.set_attrs("clust", {k: 0 if v is np.NaN else v for k, v in rs.items()})
        return graph

    def choice(self, graph, sz):
        nodes = graph.get_ids(stable=True)
        probs = graph.get_attributes("clust") ** self.param[0] + self.param[1]

        nodes, probs = nodes[probs > 0], probs[probs > 0]
        if probs.size == 0:
            return None
        probs = probs / np.sum(probs)
        return np.random.choice(nodes, sz, replace=False, p=probs)

    def clean(self, graph):
        graph = self.prep(graph)
        if self.param[1] == 0:
            return []

        dels = graph.get_ids(stable=True)[(graph.get_attributes("clust") == 0)]
        return dels



import numpy as np

from pge.model.growth.pa import PADiGrowth
from pge.model.spreading.spreader.basic import SpreadingModel


class PADirSpread(PADiGrowth, SpreadingModel):
    def __init__(self, graph, schema, deg, model_param, rs):
        super().__init__(graph, schema, deg, model_param)
        self.received = {str(node): 1 for node in graph.get_ids()}
        self.n = 0
        self.rs = rs

    def stop(self):
        return self.rs == self.n - self.gr.size()

    def new_node_add(self, graph, to, tp, attrs):
        super().new_node_add(graph, to, tp, attrs)

        self.received.update({to: int(tp == 2)})
        self.n += int(tp == 2)

    def new_edge_add(self, gr, attrs):
        node1, node2 = self.choice(gr, 1, tp="out")[0], self.choice(gr, 1, tp="in")[0]
        gr.add_edge(node1, node2, {attrs[0]: attrs[1] + 1})

        if self.received[str(node1)] == 1:
            self.n += int(self.received[str(node2)] != 1)
            self.received.update({str(node2): 1})

    def new_load(self, gr):
        self.n = self.gr.size()
        return gr

    def iteration_bunch(self, num_iter):
        times = []

        for _ in np.arange(num_iter):
            gr = self.proceed(self.rs*10000)
            times.append(gr.size_edge()-self.gr.size_edge())
        return times




import time
import numpy as np


class Spreading:
    @staticmethod
    def get_modeling_spreading_time_full_min(
        timer, model, num_iter, mx=None, prnt=False
    ):
        times = model.iteration_bunch(num_iter, mx=mx, prnt=prnt)
        df_times = [timer.get_times(int(mx_)) for mx_ in np.max(times, axis=1)]

        for i in np.arange(times.shape[0]):
            for j in np.arange(times.shape[1]):
                times[i, j] = np.sum(df_times[i][: int(times[i, j])])
        return {
            model.ids[i]: np.min(times[:, i]) for i in np.arange(model.graph.size())
        }

    @staticmethod
    def get_modeling_spreading_tick_full_min(model, num_iter, mx=None, prnt=False):
        times = model.iteration_bunch(num_iter, mx=mx, prnt=prnt)
        return {
            model.ids[i]: np.min(times[:, i]) for i in np.arange(model.graph.size())
        }

    @staticmethod
    def get_modeling_spreading_until_tick(model, num_iter, tick, prnt=False):
        return model.iteration_bunch_comm(num_iter, tick, prnt=prnt)

    @staticmethod
    def get_modeling_spreading_time_complex(timer, model, num_iter, comm, prnt=False):
        times = []
        comm_times = []

        for _ in np.arange(num_iter):
            k, n = model.iteration_bunch_complex(comm)
            df_times = timer.get_times(n)
            times.append(np.sum(df_times))
            comm_times.append(np.sum(df_times[:k]))
            if prnt:
                print("iteration", _)
        return comm_times, times

    @staticmethod
    def get_modeling_spreading_time_min_timer(model, num_iter):
        mx = None

        res = []
        for _ in np.arange(num_iter):
            toc = time.perf_counter()
            times = model.iteration_bunch(1, mx=mx, prnt=False)
            mx = np.max(times)
            res.append(time.perf_counter() - toc)
        return res

    @staticmethod
    def get_modeling_spreading_time_min_best_view(timer, model, size=None):
        nses = model.iteration_timer(size)
        times = timer.get_times(len(nses))

        ts = [np.sum(times[: i + 1]) for i in np.arange(len(nses))]
        speed = np.divide(ts, nses)
        return ts, speed, nses



import numpy as np

from pge.wlk.walker import RandomWalk
from rd_pr.simple import uniform


class RandomReWalk(RandomWalk):
    def generate(self, n, st=None, c=0.15, end_start=False):
        steps = uniform(n)
        res = []

        if st is None:
            res.append(np.random.choice(self.gr.get_ids(), size=1))
        else:
            res.append(st)

        for i in np.arange(1, n):
            if self.gr.count_out_degree(res[i-1]) == 0:
                if end_start:
                    res.append(res[0])
                else:
                    if i >= 2:
                        res.append(res[-2])
                    else:
                        res.append(res[0])

            if steps[i] < c:
                res.append(res[0])
            else:
                res.append(np.random.choice(self.gr.get_out_degrees(res[-1])))
        return res



import numpy as np
from random import choice, randint

from pge.model.spreading.spreader.basic import SpreadingModel


class Burst(SpreadingModel):
    def __init__(self, graph, alpha):
        super().__init__(graph)
        self.alpha = alpha

    def set_alpha(self, alpha):
        self.alpha = alpha

    def iteration(self):
        ind = randint(0, self.ids.size - 1)
        node = self.ids[ind]

        if self.graph.directed():
            others = self.graph.get_in_degrees(node)
        else:
            others = self.graph.get_degrees(node)

        if others.size != 0:
            u = choice(others)
            old = list(self.status[node])
            self.status[node].update(self.status[u])

            if self.received[ind] != len(self.status[node]):
                self.received[ind] = len(self.status[node])
                nws = np.setdiff1d(list(self.status[node]), old)
                self.messes[np.isin(self.ids, nws)] += 1



import numpy as np

from joblib import Parallel, delayed

from dtg.stationary.giratis import stat
from pge.ranks.extrem_onl import NodeExInfo


class LineEx(NodeExInfo):
    @staticmethod
    def get_exes(gr, root, params):
        paths, di = gr.get_short_line_paths(root)
        res = []

        for params_ in params:
            xs, lv = gr.get_ln_attrs(params_[0], paths)
            xs2 = []
            for xs_ in xs:
                xs2 = np.append(xs2, xs_)
            ex1, ex2 = LineEx.estimate(xs, lv)
            res.append((ex1, ex2, params_[1], stat(xs2)))
        return res

    @staticmethod
    def estimate(xs, lv):
        exes = []
        ms = []

        for t1 in [True, False]:
            res = np.array(
                Parallel(n_jobs=2, backend="multiprocessing")(
                    delayed(LineEx.ex_estimate)(xs, u, t1) for u in lv
                )
            )
            res_, us_ = res[res < 1], lv[res < 1]
            if res_.size > 0:
                ex = res_[-1]
                u = us_[-1]
            else:
                ex = res[-1]
                u = lv[-1]

            exes.append(ex)
            ms.append(
                np.sum([np.sum(xs_ <= u) for xs_ in xs])
                / np.sum([xs_.size for xs_ in xs])
            )

        return (exes[0], ms[0]), (exes[1], ms[1])

    @staticmethod
    def ex_estimate(xs, lv, t1):
        ts = np.array([])

        for xs_ in xs:
            if t1:
                ts_ = np.where(xs_ > lv)[0]
            else:
                ts_ = np.where(xs_ <= lv)[0]

            if ts_.size > 1:
                ts = np.append(ts, np.diff(ts_))

        if ts.size == 0:
            return 1

        if np.max(ts) > 2:
            ex = min(
                [
                    1,
                    2
                    * np.sum(ts - 1) ** 2
                    / (ts.size * np.sum(np.multiply(ts - 1, ts - 2))),
                ]
            )
        else:
            ex = min([1, 2 * np.sum(ts) ** 2 / (ts.size * np.sum(ts ** 2))])
        return ex



import leidenalg
import igraph as ig
import networkx as nx
import numpy as np


def modularity(partition, graph):
    return nx.algorithms.community.modularity(graph.get_nx_graph(), partition)


def find_best_partition(graph):
    if graph.directed():
        g = ig.Graph.from_networkx(graph.get_nx_graph())
        res_p = leidenalg.find_partition(g, leidenalg.ModularityVertexPartition).membership

        res = []
        for com in np.unique(res_p):
            res.append(set([g.vs[node]["_nx_name"] for node in np.arange(len(res_p)) if res_p[node] == com]))
        return res
    else:
        return nx.algorithms.community.greedy_modularity_communities(graph.get_nx_graph())



import networkx as nx


def estimate_rank(graph, tp, pers=None, c=0.85, max_iter=100, eps=0.000001, w=None):
    if pers is None:
        tel = {node: 1 / graph.size() for node in graph.get_ids()}
    else:
        tel = graph.get_attrs_dict(pers)
    res = nx.pagerank(graph.get_nx_graph(), alpha=c, personalization=tel,
                      max_iter=max_iter, tol=eps, weight=w)
    res = {i: res.get(i) * graph.size() for i in res}
    graph.set_attrs(tp, res)


def esp(graph, tp, frm, per="tel", c=0.85, w=None):
    nodes = graph.get_ids()
    graph.set_attrs(tp, 0)

    for node in nodes:
        rank = (1 - c) * graph.get_attr(node, per)
        k = graph.get_in_degrees(node, w=w)

        for n in k:
            t = graph.count_out_degree(n)
            if t != 0:
                t = c * graph.get_attr(n, frm) / t

            if tp == "max":
                rank = max(rank, t)
            else:
                rank = min(rank, t)
        graph.set_attr(node, tp, rank)


def spec_estimate_rank(graph, tp, to, pers=None, c=0.85, max_iter=100, eps=0.000001, w=None):
    estimate_rank(graph, tp, pers=pers, c=c, max_iter=max_iter, eps=eps, w=w)

    nodes = graph.get_nodes_with(to[0], to[1])

    res = {}
    for node in nodes:
        if pers is None:
            rank = (1 - c) / graph.size()
        else:
            rank = (1 - c) * graph.get_attr(node, pers)

        for n in graph.get_in_degrees(node):
            t = graph.count_out_degree(n)
            if t != 0:
                rank += c * graph.get_attr(n, tp) / t

        res.update({node: rank})

    for node in nodes:
        graph.set_attr(node, tp, res.get(node))



class SpreadingModel(object):
    def __init__(self, graph):
        self.graph = graph
        self.initial_status = {}

    def set_initial_status(self, configuration):
        return

    def init(self):
        return

    def iteration_bunch(self, num_iter):
        return

    def iteration_bunch_comm(self, num_iter, tick):
        return

    def iteration_bunch_complex(self, nodes):
        return

    def iteration_timer(self):
        return

    def finish(self, nodes=None):
        return True

    def iteration(self):
        return None, True



import networkx as nx
import numpy as np
import pandas as pd

from dtg.tail.estimate.hill import HillEstimator
from dtg.tail.mse import boot_estimate
from pge.cluster.modularity import best_partition_to
from pge.model.growth.basic import BDiGrowth
from pge.ranks.rank import estimate_rank


class PAСomDiGrowth(BDiGrowth):
    def __init__(self, graph, schema, deg, model_param):
        super().__init__(graph, schema, deg, model_param)
        self.alps = []
        self.mods = []
        self.cur_alps = []
        self.frm = "com"

    def choice(self, graph, sz, tp="in"):
        coms = np.unique(graph.get_attributes(self.frm))

        nodes = []
        for i, com in enumerate(coms):
            if self.cur_alps[i] == self.alps[-1]:
                nodes = np.append(nodes, graph.get_nodes_with(self.frm, com))
        if tp == "in":
            probs = np.array([graph.count_in_degree(node) for node in nodes]) + self.param[0]
        else:
            probs = np.array([graph.count_in_degree(node) for node in nodes]) + self.param[1]
        probs = probs / np.sum(probs)
        return np.random.choice(nodes, sz, replace=False, p=probs)

    def prep(self, graph):
        estimate_rank(graph, "one", pers=None)
        coms = best_partition_to(graph.clean_copy())
        graph.set_attrs(self.frm, coms[0])
        self.mods.append(coms[1])

        self.cur_alps = []
        for com in set(coms[0].values()):
            ones = graph.get_attributes("one", graph.get_nodes_with(self.frm, com))
            self.cur_alps.append((
                boot_estimate(
                    HillEstimator,
                    ones,
                    1 / 2,
                    2 / 3,
                    30,
                    speed=False,
                )[0]
            ))
            if self.cur_alps[-1] is None:
                self.cur_alps[-1] = np.infty
            else:
                self.cur_alps[-1] = 1 / self.cur_alps[-1]
        self.alps.append(self.cur_alps[np.argmin(self.cur_alps)])

        graph.set_attrs(
            "dg_in", {node: graph.count_in_degree(node) for node in graph.get_ids()}
        )
        graph.set_attrs(
            "dg_out", {node: graph.count_out_degree(node) for node in graph.get_ids()}
        )
        return graph

    def save(self, gr, to):
        nx.write_graphml(gr, to + ".graphml")
        prd = pd.DataFrame({"alp": self.alps, "mods": self.mods})
        prd.to_csv(to + ".csv")



import numpy as np


class PoisTimer:
    def __init__(self, lamb):
        self.lamb = lamb

    def get_times(self, n):
        return np.random.poisson(self.lamb, n)


class ExpTimer:
    def __init__(self, lamb):
        self.lamb = lamb

    def get_times(self, n):
        return np.random.exponential(self.lamb, n)



from setuptools import setup, find_packages


version = "1.0.9.3"

setup(
    name="pge",
    version=version,
    description="Graph modeling, EI estimation and analysis",
    author="Maxim Ryzhov",
    author_email="ryzhov@phystech.edu",
    license="MIT",
    packages=find_packages(),
    url="https://gitlab.com/BoBeni/PGEstimate",
    install_requires=["numpy", "scipy", "networkx", "pandas", "scipy", "mcg"],
    zip_safe=False,
)





def per_node_estimation(i, gr, frm, to, templ, c, entity, st_pl, typ, view, crt, w, full=200):
    ids, exes, clust, bls = templ.get_ex(gr, i, frm, c, entity=entity, stab=st_pl, typ=typ, view=view, criterion=crt, w=w, full=full)

    if ids.size == 0:
        print("-")
        gr.set_attr(i, to, 1)
        gr.set_attr(i, to + "-cls", 1)
        gr.set_attr(i, to + "-blk", 1)
        gr.set_attr(i, to + "-sz", 1)
    else:
        gr.set_attr(i, to, exes[-1])
        print(i, ids.size-1, exes[-1], bls[-1])

        gr.set_attr(i, to + "-cls", clust[-1])
        gr.set_attr(i, to + "-blk", bls[-1])
        gr.set_attr(i, to + "-sz", ids.size)
    return ids


def ex_estimate(gr, frm, to, templ, c, entity, crt, st_pl=50, typ="sl", view=5, w=False):
    for n in gr.get_ids():
        per_node_estimation(n, gr, frm, to, templ, c, entity, st_pl, typ, view, crt, w)


class NodeExInfo:
    @staticmethod
    def get_ex(gr, roots, frm):
        return None

    @staticmethod
    def double_value():
        return False


class CommExInfo:
    @staticmethod
    def get_ex(gr, roots, frm):
        return None

    @staticmethod
    def double_value():
        return False



import numpy as np

from pge.wlk.walker import RandomWalk
from rd_pr.simple import uniform


class RandomJumpingWalk(RandomWalk):
    def generate(self, n, st=None, c=0.15, end_start=False):
        steps = uniform(n)

        if st is None:
            news = np.random.choice(self.gr.get_ids(), size = np.sum(steps < c) + 1)
            res = [news[0]]
            news = news[1:]
        else:
            news = np.random.choice(self.gr.get_ids(), size= np.sum(steps < c))
            res = [st]

        count = 0
        for i in np.arange(1, n):
            if self.gr.count_out_degree(res[-1]) == 0:
                if end_start:
                    res.append(res[0])
                else:
                    if i >= 2:
                        res.append(res[-2])
                    else:
                        res.append(res[0])

            if c > steps[i]:
                res.append(news[count])
                count += 1
            else:
                res.append(np.random.choice(self.gr.get_out_degrees(res[-1])))

        return np.array(res)


